unit sensmain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, stypes;

type
  TFmSensMain = class(TForm)
    EdListFile: TEdit;
    LblListFile: TLabel;
    MmInstruct: TMemo;
    EdOutputFile: TEdit;
    LblOutputFile: TLabel;
    BtnRunSens: TButton;
    BtnCloseSens: TButton;
    DlgOpenList: TOpenDialog;
    DlgSaveSensOutput: TSaveDialog;
    LblBatchFile: TLabel;
    EdBatchFile: TEdit;
    LblMeasFile: TLabel;
    EdMeasFile: TEdit;
    DlgOpenMeasFile: TOpenDialog;
    procedure ChooseListFile(Sender: TObject);
    procedure LblListFileClick(Sender: TObject);
    procedure BtnRunSensClick(Sender: TObject);
    procedure BtnCloseSensClick(Sender: TObject);
    procedure ChooseSensOutputFile(Sender: TObject);
    procedure LblBatchFileClick(Sender: TObject);
    procedure GetMeasurements;
    procedure GetSimResults;
    procedure LblMeasFileClick(Sender: TObject);
    procedure ChooseMeasFile(Sender: TObject);
    procedure ParseMeasData(datastring:string; var MeasuredData:TTreatResults);
    function GetTreatment(filename:string):string;
    procedure FormCreate(Sender: TObject);
    function NumberofMeasurements(treatment:string):integer;
    procedure CalculateResiduals;
    procedure UpdateForm;
    procedure ChooseBatchFile(Sender:TObject);

  private
    { Private declarations }

  public
    { Public declarations }
    ResidCL, ResidCW, ResidCR, ResidNL, ResidNW, ResidNR, ResidCT,
       ResidNT, ResidNEP:double;
    MeasData, SimData, Residuals:ResultArray;
    ListFileName, SensOutputFile, MeasFilename: string;
    NumTests:integer;
  end;

var
  FmSensMain: TFmSensMain;

implementation

{$R *.DFM}
uses frontend, calculate, fileio, batchmain, display;

procedure TFmSensMain.LblListFileClick(Sender: TObject);
begin
 ChooseListFile(FmSensMain);
 EdListFile.Text := listfilename;
end;

procedure TFmSensMain.ChooseListFile(Sender: TObject);
begin
// If the user typed directly in the edit box
 if Sender is TEdit then
   if not FileExists(ListFileName) then
    begin
     MessageDlg('Invalid List File: ' + ListFileName + '. Please reenter.',
                  mtError, [mbOK], 0);
     ListFileName := '';  // Set the filename to no value
     EdListFile.text := ListFileName;
    end
   else
     ListFileName:=EdListFile.text  // Set the list file
 else  // User clicked on the label
  begin   // Show the open file dialog
    // First set the default filename to the current file
   DlgOpenList.filename := ListFileName;
    // Show the dialog and if the user clicks OK, set the new filename
   if DlgOpenList.execute then
    begin
      ListFileName := DlgOpenList.filename;
    end;
  end;
 UpdateForm;
end;

procedure TFmSensMain.BtnRunSensClick(Sender: TObject);
var
 initialparamfile, initialdriverfile, sensparameter, Path, TestStatus:string;
 newparamvalue, temp:double;
 arraynum:integer;
begin
 BtnRunSens.Enabled := False;
 BtnCloseSens.Enabled := False;
 OpenListFile(ListFileName);
 if not ReadListFile(initialparamfile,initialdriverfile,sensparameter,newparamvalue)
   then raise Exception.Create('Error reading List File');
 OpenSensOutFile(SensOutputFile);
 try
  repeat
    try
     paramfilename := initialparamfile;
     ReadParamFile(paramfilename,ModelDef.numparam,par,ModelDef.numstate,stat,temp);
//     MainForm.ChooseParamFile(Sender); // Read in initial parameters and states
// Calculate a steady state with new param value and save as senstest.par
     driverfilename := initialdriverfile;
     FmCalculate.FormShow(Sender);     // Calculate a steadystate

     arraynum := FmCalculate.GetArrayIndex(vtparameter,sensparameter);
     FmCalculate.ChangeValue(vtparameter, arraynum, newparamvalue, True);
     FmCalculate.BtnOKClick(Sender);
     FmCalculate.Hide;
     Path := ExtractFilePath(initialparamfile);
     paramfilename := Path + 'senstest.par';
     MainForm.MISaveParamClick(Sender);

// Run model runs listed in the batch file
     FmBatchMain.BtnRunBatchClick(Sender);

// Get the results of the test runs and calculate the residuals
     GetMeasurements;
     GetSimResults;
     CalculateResiduals;

     TestStatus := 'Test Complete';
     WriteSensOutFile(initialparamfile, sensparameter, newparamvalue,
      TestStatus, ResidCL, ResidCW, ResidCR, ResidNL, ResidNW, ResidNR,
      ResidCT, ResidNT);
    except
     on E: Exception do
      begin
       E.Message := 'Error running sensitivity test. ' + E.Message +
          '. Test Aborted.';
       TestStatus := E.Message;
       ResidCL := 0; ResidCW := 0; ResidCR := 0; ResidCT := 0;
       ResidNL := 0; ResidNW := 0; ResidNR := 0; ResidNT := 0;
       WriteSensOutFile(initialparamfile, sensparameter, newparamvalue,
        TestStatus, ResidCL, ResidCW, ResidCR, ResidNL, ResidNW, ResidNR,
        ResidCT, ResidNT);
//       raise;
      end;
    end;
  until ReadListFile(initialparamfile,initialdriverfile,sensparameter,
               newparamvalue) = False;
 finally
  BtnRunSens.Enabled := True;
  BtnCloseSens.Enabled := True;
  CloseListFile;
  CloseSensOutFile;
 end;
end;

procedure TFmSensMain.BtnCloseSensClick(Sender: TObject);
begin
  FmSensMain.Close;
end;

procedure TFmSensMain.ChooseSensOutputFile(Sender: TObject);
begin
   // If the user typed directly in the edit box
 if Sender is TEdit then
  begin
    if FileExists(SensOutputFile) then  // If the file already exists
      // Confirm that the user wants to overwrite the existing file.
       if MessageDlg('Overwrite existing output file, ' + SensOutputFile + '?',
                  mtConfirmation, [mbYes,mbNo], 0) = mrNo then
         SensOutputFile := ''  // Set the outfile to no value
       else
         SensOutputFile := EdOutputFile.text  // Set the output file
  end
 else   // User used the menu or clicked on the Output label
  begin  // Show the Save file dialog
    // First set the default filename to the current output file.
   DlgSaveSensOutput.filename := SensOutputFile;
   // Show the dialog and if the user chooses OK, set the outfilename to the new file.
   if DlgSaveSensOutput.execute then
         SensOutputFile := DlgSaveSensOutput.filename;
   EdOutputFile.text := SensOutputFile;
  end;
 UpdateForm;
end;

procedure TFmSensMain.ChooseBatchFile(Sender: TObject);
begin
 FmBatchMain.DlgOpenBatch.Options := FmBatchMain.DlgOpenBatch.Options +
                                      [ofNoChangeDir];
 FmBatchMain.ChooseBatchFile(Sender);
end;

procedure TFmSensMain.LblBatchFileClick(Sender: TObject);
begin
 ChooseBatchFile(Sender);
 UpdateForm;
end;

procedure TFmSensMain.GetMeasurements;
var
 tempstring: string;
 i: integer;
begin
 OpenMeasFile(measFilename);
 try
   if not ReadMeasFile(tempstring)then
     raise Exception.Create('Error reading measured data.');
   i := 1;
   repeat
     ParseMeasData(tempstring,MeasData[i]);
     i := i + 1;
     if i > maxresults then
       raise Exception.Create('Too many measured data. ' +
           ' Increase maxresults in stypes.pas and repeat sensitivity run.');
   until ReadMeasFile(tempstring) = false;
 finally
   CloseMeasFile;
 end;
end;

procedure TFmSensMain.ParseMeasData(datastring:string; var
                                      MeasuredData:TTreatResults);
var
 numbegin, numend: integer;
 tempstring:string;
begin
 RemoveSpaces(datastring);

 numbegin := 1;
 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.Treatment := tempstring;

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.Year := Strtoint(tempstring);

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.States.Cl.value := strtofloat(tempstring);

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.States.Cw.value := strtofloat(tempstring);

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.States.Cr.value := strtofloat(tempstring);

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.States.Nl.value := strtofloat(tempstring);

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.States.Nw.value := strtofloat(tempstring);

 numend := pos(',',datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.States.Nr.value := strtofloat(tempstring);

 numend :=  length(datastring);
 tempstring := copy(datastring,numbegin,numend-numbegin+1);
 delete(datastring,numbegin,numend);
 delete(tempstring,pos(',',tempstring),1);
 MeasuredData.NEP := strtofloat(tempstring);
end;

procedure TFmSensMain.GetSimResults;
var
 j, testnum, rownum, numtreat, FoliageCol, NEPcol, NPPcol:integer;
 treatment:string;

function GetColumnNumber(name:string):integer;
var
 i,colnum:integer;
 tempstring:string;
begin
 i := 1;
 colnum := 0;
 repeat
   tempstring := FmDisplayData.SgModelOutput.Cells[i,1];
   tempstring := trim(tempstring);
   if tempstring = name then colnum := i;
   i := i + 1;
 until (colnum <> 0) or (i > FmDisplayData.SgModelOutput.Colcount - 1);
 result := colnum;
end;

begin
 OpenBatchFile(BatchFilename);
 try
   ReadBatchFile(paramfilename,driverfilename,outfilename,time_start,
                           time_stop,time_step);
   testnum := 1;
   FoliageCol := GetColumnNumber('Total Foliage Carbon');
   NEPCol := GetColumnNumber('Net Ecosys Prod.');
   NPPCol := GetColumnNumber('Net Plant Prod.');
   repeat
     if testnum > maxresults then
       raise Exception.Create('Too many simulated data.');
     FmDisplayData.Filename := outfilename;
     FmDisplayData.UpdateStringGrid;
     treatment := GetTreatment(outfilename);
     numTreat := NumberofMeasurements(treatment);
     if numTreat = 0 then
       raise Exception.Create('No measured data found for ' +
                 treatment + ' treatment. ');
     for j := 1 to numTreat do
      begin
       rownum := MeasData[testnum].year + 3;
       SimData[testnum].States.Cl.value :=               // col, row
            strtofloat(FmDisplayData.SgModelOutput.Cells[FoliageCol,rownum]);
       SimData[testnum].States.Cw.value :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[FoliageCol+1,rownum]);
       SimData[testnum].States.Cr.value :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[FoliageCol+2,rownum]);
       SimData[testnum].States.Nl.value :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[FoliageCol+3,rownum]);
       SimData[testnum].States.Nw.value :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[FoliageCol+4,rownum]);
       SimData[testnum].States.Nr.value :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[FoliageCol+5,rownum]);
       SimData[testnum].NEP :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[NEPcol,rownum]);
       SimData[testnum].NPP :=
            strtofloat(FmDisplayData.SgModelOutput.Cells[NPPcol,rownum]);
       testnum := testnum + 1;
      end;
   until ReadBatchFile(paramfilename,driverfilename,outfilename,
                           time_start,time_stop,time_step) = False;
   NumTests := testnum - 1;
 finally
  CloseBatchFile;
 end;
end;

procedure TFmSensMain.CalculateResiduals;
var
 i:integer;
begin
 ResidCL := 0;
 ResidCW := 0;
 ResidCR := 0;
 ResidNL := 0;
 ResidNW := 0;
 ResidNR := 0;
 ResidNEP := 0;
 for i := 1 to NumTests do
  begin
    Residuals[i].Treatment := MeasData[i].Treatment;
    Residuals[i].Year := MeasData[i].Year;
    if MeasData[i].States.Cl.value = 0 then
      Residuals[i].States.Cl.value := 0
    else
      Residuals[i].States.Cl.value := abs(MeasData[i].States.Cl.value -
       SimData[i].States.Cl.value)/MeasData[i].States.Cl.value;
    if MeasData[i].States.Cw.value = 0 then
      Residuals[i].States.Cw.value := 0
    else
      Residuals[i].States.Cw.value := abs(MeasData[i].States.Cw.value -
       SimData[i].States.Cw.value)/MeasData[i].States.Cw.value;
//    if MeasData[i].States.Cr.value = 0 then
//      Residuals[i].States.Cr.value := 0
//    else
//      Residuals[i].States.Cr.value := abs(MeasData[i].States.Cr.value -
//       SimData[i].States.Cr.value)/MeasData[i].States.Cr.value;
    if MeasData[i].States.Nl.value = 0 then
      Residuals[i].States.Nl.value := 0
    else
      Residuals[i].States.Nl.value := abs(MeasData[i].States.Nl.value -
       SimData[i].States.Nl.value)/MeasData[i].States.Nl.value;
    if MeasData[i].States.Nw.value = 0 then
      Residuals[i].States.Nw.value := 0
    else
      Residuals[i].States.Nw.value := abs(MeasData[i].States.Nw.value -
       SimData[i].States.Nw.value)/MeasData[i].States.Nw.value;
    if MeasData[i].States.Nr.value = 0 then
      Residuals[i].States.Nr.value := 0
    else
      Residuals[i].States.Nr.value := abs(MeasData[i].States.Nr.value -
       SimData[i].States.Nr.value)/MeasData[i].States.Nr.value;
    if MeasData[i].NEP = 0 then
      Residuals[i].NEP := 0
    else
      Residuals[i].NEP := abs(MeasData[i].NEP -
       SimData[i].NEP)/SimData[i].NPP;
    ResidCL := ResidCL + Residuals[i].States.Cl.value;
    ResidCW := ResidCW + Residuals[i].States.Cw.value;
    ResidCR := ResidCR + Residuals[i].States.Cr.value;
    ResidNL := ResidNL + Residuals[i].States.Nl.value;
    ResidNW := ResidNW + Residuals[i].States.Nw.value;
    ResidNR := ResidNR + Residuals[i].States.Nr.value;
    ResidNEP := ResidNEP + Residuals[i].NEP;
  end;
 ResidCT := ResidCL + ResidCW + ResidCR;
 ResidNT := ResidNL + ResidNW + ResidNR + ResidNEP;
end;

procedure TFmSensMain.LblMeasFileClick(Sender: TObject);
begin                                       
 ChooseMeasFile(Sender);
 UpdateForm;
end;

procedure TFmSensMain.ChooseMeasFile(Sender: TObject);
var
 ThisEdit:TEdit;
begin
// If the user typed directly in the edit box
 if Sender is TEdit then
  begin
   ThisEdit := Sender as TEdit;
   Measfilename := ThisEdit.text;
   if not FileExists(Measfilename) then
    begin
     MessageDlg('Invalid Data File: ' + Measfilename + '. Please reenter.',
                  mtError, [mbOK], 0);
     Measfilename := '';  // Set the filename to no value
     ThisEdit.text := Measfilename;
    end;
  end
 else  // User clicked on the label
  begin   // Show the open file dialog
   ThisEdit := (Sender as Tlabel).focuscontrol as TEdit;
    // First set the default filename to the current file
   DlgOpenMeasFile.filename := Measfilename;
    // Show the dialog and if the user clicks OK, set the new filename
   if DlgOpenMeasFile.execute then
    begin
      Measfilename := DlgOpenMeasFile.filename;
      ThisEdit.text := measfilename;
    end;
  end;
end;

function TFmSensMain.GetTreatment(filename:string):string;
var
 tempstring:string;
begin
 lowercase(filename);
 if pos('ct',filename)<> 0 then tempstring := 'CT'
 else if pos('ghf', filename) <> 0 then tempstring := 'GHF'
 else if pos('np',filename) <> 0 then tempstring := 'NP'
 else if pos('gh',filename) <> 0 then tempstring := 'GH'
 else if pos('sh',filename) <> 0 then tempstring := 'SH'
 else if pos('co2',filename) <> 0 then tempstring := 'CO2'
 else tempstring := 'Error';
 GetTreatment := tempstring;
end;

procedure TFmSensMain.FormCreate(Sender: TObject);
var
 i:integer;
begin
 for i := 1 to maxresults do
  begin
    SimData[i].States.cl := stat[1];
    SimData[i].States.cw := stat[2];
    SimData[i].States.cr := stat[3];
    SimData[i].States.nl := stat[4];
    SimData[i].States.nw := stat[5];
    SimData[i].States.nr := stat[6];
  end;
 Residuals := SimData;
end;

function TFmSensMain.NumberofMeasurements(treatment:string):integer;
var
 i, count:integer;
begin
 count := 0;
 for i := 1 to maxresults do
  begin
    if MeasData[i].treatment = treatment then count := count + 1;
  end;
 NumberofMeasurements := count;
end;

procedure TFmSensMain.UpdateForm;
begin
 if (ListFileName <> '') and (SensOutputFile <> '') and (MeasFilename <> '')
    and (BatchFilename <> '') then
   begin
     BtnRunSens.Enabled := True;
     FmSensMain.ActiveControl := BtnRunSens;
   end;
end;

end.


